import pytest
import torch

from torchmeta.modules import MetaLinear
from torchmeta.utils.gradient_based import gradient_update_parameters

@pytest.fixture
def model():
    model = MetaLinear(3, 1, bias=False)
    model.weight.data = torch.tensor([[2., 3., 5.]])
    return model


def test_update_parameters(model):
    """
    The loss function (with respect to the weights of the model w) is defined as
        f(w) = 0.5 * (1 * w_1 + 2 * w_2 + 3 * w_3) ** 2
    with w = [2, 3, 5].

    The gradient of the function f with respect to w, and evaluated
    at w = [2, 3, 5], is:
        df / dw_1 = 1 * (1 * w_1 + 2 * w_2 + 3 * w_3) = 23
        df / dw_2 = 2 * (1 * w_1 + 2 * w_2 + 3 * w_3) = 46
        df / dw_3 = 3 * (1 * w_1 + 2 * w_2 + 3 * w_3) = 69

    The updated parameter w' is then given by one step of gradient descent,
    with step size 0.5:
        w'_1 = w_1 - 0.5 * df / dw_1 = 2 - 0.5 * 23 = -9.5
        w'_2 = w_2 - 0.5 * df / dw_2 = 3 - 0.5 * 46 = -20
        w'_3 = w_3 - 0.5 * df / dw_3 = 5 - 0.5 * 68 = -29.5
    """
    train_inputs = torch.tensor([[1., 2., 3.]])
    train_loss = 0.5 * (model(train_inputs) ** 2)

    params = gradient_update_parameters(model,
                                        train_loss,
                                        params=None,
                                        step_size=0.5,
                                        first_order=False)

    assert train_loss.item() == 264.5
    assert list(params.keys()) == ['weight']
    assert torch.all(params['weight'].data == torch.tensor([[-9.5, -20., -29.5]]))

    """
    The new loss function (still with respect to the weights of the model w) is
    defined as:
        g(w) = 0.5 * (4 * w'_1 + 5 * w'_2 + 6 * w'_3) ** 2
             = 0.5 * (4 * (w_1 - 0.5 * df / dw_1)
                    + 5 * (w_2 - 0.5 * df / dw_2)
                    + 6 * (w_3 - 0.5 * df / dw_3)) ** 2
             = 0.5 * (4 * (w_1 - 0.5 * 1 * (1 * w_1 + 2 * w_2 + 3 * w_3))
                    + 5 * (w_2 - 0.5 * 2 * (1 * w_1 + 2 * w_2 + 3 * w_3))
                    + 6 * (w_3 - 0.5 * 3 * (1 * w_1 + 2 * w_2 + 3 * w_3))) ** 2
             = 0.5 * ((4 - 4 * 0.5 - 5 * 1.0 - 6 * 1.5) * w_1
                    + (5 - 4 * 1.0 - 5 * 2.0 - 6 * 3.0) * w_2
                    + (6 - 4 * 1.5 - 5 * 3.0 - 6 * 4.5) * w_3) ** 2
             = 0.5 * (-12 * w_1 - 27 * w_2 - 42 * w_3) ** 2

    Therefore the gradient of the function g with respect to w (and evaluated
    at w = [2, 3, 5]) is:
        dg / dw_1 = -12 * (-12 * w_1 - 27 * w_2 - 42 * w_3) =  3780
        dg / dw_2 = -27 * (-12 * w_1 - 27 * w_2 - 42 * w_3) =  8505
        dg / dw_3 = -42 * (-12 * w_1 - 27 * w_2 - 42 * w_3) = 13230
    """
    test_inputs = torch.tensor([[4., 5., 6.]])
    test_loss = 0.5 * (model(test_inputs, params=params) ** 2)

    grads = torch.autograd.grad(test_loss, model.parameters())

    assert test_loss.item() == 49612.5
    assert len(grads) == 1
    assert torch.all(grads[0].data == torch.tensor([[3780., 8505., 13230.]]))


def test_update_parameters_first_order(model):
    """
    The loss function (with respect to the weights of the model w) is defined as
        f(w) = 0.5 * (4 * w_1 + 5 * w_2 + 6 * w_3) ** 2
    with w = [2, 3, 5].

    The gradient of the function f with respect to w, and evaluated
    at w = [2, 3, 5] is:
        df / dw_1 = 4 * (4 * w_1 + 5 * w_2 + 6 * w_3) = 212
        df / dw_2 = 5 * (4 * w_1 + 5 * w_2 + 6 * w_3) = 265
        df / dw_3 = 6 * (4 * w_1 + 5 * w_2 + 6 * w_3) = 318

    The updated parameter w' is then given by one step of gradient descent,
    with step size 0.5:
        w'_1 = w_1 - 0.5 * df / dw_1 = 2 - 0.5 * 212 = -104
        w'_2 = w_2 - 0.5 * df / dw_2 = 3 - 0.5 * 265 = -129.5
        w'_3 = w_3 - 0.5 * df / dw_3 = 5 - 0.5 * 318 = -154
    """
    train_inputs = torch.tensor([[4., 5., 6.]])
    train_loss = 0.5 * (model(train_inputs) ** 2)

    params = gradient_update_parameters(model,
                                        train_loss,
                                        params=None,
                                        step_size=0.5,
                                        first_order=True)

    assert train_loss.item() == 1404.5
    assert list(params.keys()) == ['weight']
    assert torch.all(params['weight'].data == torch.tensor([[-104., -129.5, -154.]]))

    """
    The new loss function (still with respect to the weights of the model w) is
    defined as:
        g(w) = 0.5 * (1 * w'_1 + 2 * w'_2 + 3 * w'_3) ** 2

    Since we computed w' with the first order approximation, the gradient of the
    function g with respect to w, and evaluated at w = [2, 3, 5], is:
        dg / dw_1 = 1 * (1 * w'_1 + 2 * w'_2 + 3 * w'_3) =  -825
        dg / dw_2 = 2 * (1 * w'_1 + 2 * w'_2 + 3 * w'_3) = -1650
        dg / dw_3 = 3 * (1 * w'_1 + 2 * w'_2 + 3 * w'_3) = -2475
    """
    test_inputs = torch.tensor([[1., 2., 3.]])
    test_loss = 0.5 * (model(test_inputs, params=params) ** 2)

    grads = torch.autograd.grad(test_loss, model.parameters())

    assert test_loss.item() == 340312.5
    assert len(grads) == 1
    assert torch.all(grads[0].data == torch.tensor([[-825., -1650., -2475.]]))


def test_multiple_update_parameters(model):
    """
    The loss function (with respect to the weights of the model w) is defined as
        f(w) = 0.5 * (1 * w_1 + 2 * w_2 + 3 * w_3) ** 2
    with w = [2, 3, 5].

    The gradient of f with respect to w is:
        df / dw_1 = 1 * (1 * w_1 + 2 * w_2 + 3 * w_3) = 23
        df / dw_2 = 2 * (1 * w_1 + 2 * w_2 + 3 * w_3) = 46
        df / dw_3 = 3 * (1 * w_1 + 2 * w_2 + 3 * w_3) = 69

    The updated parameters are given by:
        w'_1 = w_1 - 1. * df / dw_1 = 2 - 1. * 23 = -21
        w'_2 = w_2 - 1. * df / dw_2 = 3 - 1. * 46 = -43
        w'_3 = w_3 - 1. * df / dw_3 = 5 - 1. * 69 = -64
    """
    train_inputs = torch.tensor([[1., 2., 3.]])

    train_loss_1 = 0.5 * (model(train_inputs) ** 2)
    params_1 = gradient_update_parameters(model,
                                          train_loss_1,
                                          params=None,
                                          step_size=1.,
                                          first_order=False)

    assert train_loss_1.item() == 264.5
    assert list(params_1.keys()) == ['weight']
    assert torch.all(params_1['weight'].data == torch.tensor([[-21., -43., -64.]]))

    """
    The new loss function is defined as
        g(w') = 0.5 * (1 * w'_1 + 2 * w'_2 + 3 * w'_3) ** 2
    with w' = [-21, -43, -64].

    The gradient of g with respect to w' is:
        dg / dw'_1 = 1 * (1 * w'_1 + 2 * w'_2 + 3 * w'_3) = -299
        dg / dw'_2 = 2 * (1 * w'_1 + 2 * w'_2 + 3 * w'_3) = -598
        dg / dw'_3 = 3 * (1 * w'_1 + 2 * w'_2 + 3 * w'_3) = -897

    The updated parameters are given by:
        w''_1 = w'_1 - 1. * dg / dw'_1 = -21 - 1. * -299 = 278
        w''_2 = w'_2 - 1. * dg / dw'_2 = -43 - 1. * -598 = 555
        w''_3 = w'_3 - 1. * dg / dw'_3 = -64 - 1. * -897 = 833
    """
    train_loss_2 = 0.5 * (model(train_inputs, params=params_1) ** 2)
    params_2 = gradient_update_parameters(model,
                                          train_loss_2,
                                          params=params_1,
                                          step_size=1.,
                                          first_order=False)

    assert train_loss_2.item() == 44700.5
    assert list(params_2.keys()) == ['weight']
    assert torch.all(params_2['weight'].data == torch.tensor([[278., 555., 833.]]))

    """
    The new loss function is defined as
        h(w'') = 0.5 * (1 * w''_1 + 2 * w''_2 + 3 * w''_3) ** 2
    with w'' = [278, 555, 833].

    The gradient of h with respect to w'' is:
        dh / dw''_1 = 1 * (1 * w''_1 + 2 * w''_2 + 3 * w''_3) =  3887
        dh / dw''_2 = 2 * (1 * w''_1 + 2 * w''_2 + 3 * w''_3) =  7774
        dh / dw''_3 = 3 * (1 * w''_1 + 2 * w''_2 + 3 * w''_3) = 11661

    The updated parameters are given by:
        w'''_1 = w''_1 - 1. * dh / dw''_1 = 278 - 1. *  3887 =  -3609
        w'''_2 = w''_2 - 1. * dh / dw''_2 = 555 - 1. *  7774 =  -7219
        w'''_3 = w''_3 - 1. * dh / dw''_3 = 833 - 1. * 11661 = -10828
    """
    train_loss_3 = 0.5 * (model(train_inputs, params=params_2) ** 2)
    params_3 = gradient_update_parameters(model,
                                          train_loss_3,
                                          params=params_2,
                                          step_size=1.,
                                          first_order=False)

    assert train_loss_3.item() == 7554384.5
    assert list(params_3.keys()) == ['weight']
    assert torch.all(params_3['weight'].data == torch.tensor([[-3609., -7219., -10828.]]))

    """
    The new loss function is defined as
        l(w) = 4 * w'''_1 + 5 * w'''_2 + 6 * w'''_3
    with w = [2, 3, 5] and w''' = [-3609, -7219, -10828].

    The gradient of l with respect to w is:
        dl / dw_1 = 4 * dw'''_1 / dw_1 + 5 * dw'''_2 / dw_1 + 6 * dw'''_3 / dw_1
                  = ... =  -5020
        dl / dw_2 = 4 * dw'''_1 / dw_2 + 5 * dw'''_2 / dw_2 + 6 * dw'''_3 / dw_2
                  = ... = -10043
        dl / dw_3 = 4 * dw'''_1 / dw_3 + 5 * dw'''_2 / dw_3 + 6 * dw'''_3 / dw_3
                  = ... = -15066
    """
    test_inputs = torch.tensor([[4., 5., 6.]])
    test_loss = model(test_inputs, params=params_3)
    grads = torch.autograd.grad(test_loss, model.parameters())

    assert test_loss.item() == -115499.
    assert len(grads) == 1
    assert torch.all(grads[0].data == torch.tensor([[-5020., -10043., -15066.]]))



import warnings

from torchmeta.datasets import (Omniglot, MiniImagenet, TieredImagenet, CIFARFS,
                                FC100, CUB, DoubleMNIST, TripleMNIST, Pascal5i)
from torchmeta.transforms import Categorical, ClassSplitter, Rotation, SegmentationPairTransform
from torchvision.transforms import Compose, Resize, CenterCrop, ToTensor

__all__ = [
    'omniglot',
    'miniimagenet',
    'tieredimagenet',
    'cifar_fs',
    'fc100',
    'cub',
    'doublemnist',
    'triplemnist'
]

def helper_with_default(klass, folder, shots, ways, shuffle=True,
                        test_shots=None, seed=None, defaults={}, **kwargs):
    if 'num_classes_per_task' in kwargs:
        warnings.warn('Both arguments `ways` and `num_classes_per_task` were '
            'set in the helper function for the number of classes per task. '
            'Ignoring the argument `ways`.', stacklevel=2)
        ways = kwargs['num_classes_per_task']
    if 'transform' not in kwargs:
        kwargs['transform'] = defaults.get('transform', ToTensor())
    if 'target_transform' not in kwargs:
        kwargs['target_transform'] = defaults.get('target_transform',
                                                  Categorical(ways))
    if 'class_augmentations' not in kwargs:
        kwargs['class_augmentations'] = defaults.get('class_augmentations', None)
    if test_shots is None:
        test_shots = shots
    dataset = klass(folder, num_classes_per_task=ways, **kwargs)
    dataset = ClassSplitter(dataset, shuffle=shuffle,
        num_train_per_class=shots, num_test_per_class=test_shots)
    dataset.seed(seed)

    return dataset

def omniglot(folder, shots, ways, shuffle=True, test_shots=None,
             seed=None, **kwargs):
    """Helper function to create a meta-dataset for the Omniglot dataset.

    Parameters
    ----------
    folder : string
        Root directory where the dataset folder `omniglot` exists.

    shots : int
        Number of (training) examples per class in each task. This corresponds
        to `k` in `k-shot` classification.

    ways : int
        Number of classes per task. This corresponds to `N` in `N-way`
        classification.

    shuffle : bool (default: `True`)
        Shuffle the examples when creating the tasks.

    test_shots : int, optional
        Number of test examples per class in each task. If `None`, then the
        number of test examples is equal to the number of training examples per
        class.

    seed : int, optional
        Random seed to be used in the meta-dataset.

    kwargs
        Additional arguments passed to the `Omniglot` class.

    See also
    --------
    `datasets.Omniglot` : Meta-dataset for the Omniglot dataset.
    """
    defaults = {
        'transform': Compose([Resize(28), ToTensor()]),
        'class_augmentations': [Rotation([90, 180, 270])]
    }

    return helper_with_default(Omniglot, folder, shots, ways,
                               shuffle=shuffle, test_shots=test_shots,
                               seed=seed, defaults=defaults, **kwargs)

def miniimagenet(folder, shots, ways, shuffle=True, test_shots=None,
                 seed=None, **kwargs):
    """Helper function to create a meta-dataset for the Mini-Imagenet dataset.

    Parameters
    ----------
    folder : string
        Root directory where the dataset folder `miniimagenet` exists.

    shots : int
        Number of (training) examples per class in each task. This corresponds
        to `k` in `k-shot` classification.

    ways : int
        Number of classes per task. This corresponds to `N` in `N-way`
        classification.

    shuffle : bool (default: `True`)
        Shuffle the examples when creating the tasks.

    test_shots : int, optional
        Number of test examples per class in each task. If `None`, then the
        number of test examples is equal to the number of training examples per
        class.

    seed : int, optional
        Random seed to be used in the meta-dataset.

    kwargs
        Additional arguments passed to the `MiniImagenet` class.

    See also
    --------
    `datasets.MiniImagenet` : Meta-dataset for the Mini-Imagenet dataset.
    """
    defaults = {
        'transform': Compose([Resize(84), ToTensor()])
    }

    return helper_with_default(MiniImagenet, folder, shots, ways,
                               shuffle=shuffle, test_shots=test_shots,
                               seed=seed, defaults=defaults, **kwargs)

def tieredimagenet(folder, shots, ways, shuffle=True, test_shots=None,
                   seed=None, **kwargs):
    """Helper function to create a meta-dataset for the Tiered-Imagenet dataset.

    Parameters
    ----------
    folder : string
        Root directory where the dataset folder `tieredimagenet` exists.

    shots : int
        Number of (training) examples per class in each task. This corresponds
        to `k` in `k-shot` classification.

    ways : int
        Number of classes per task. This corresponds to `N` in `N-way`
        classification.

    shuffle : bool (default: `True`)
        Shuffle the examples when creating the tasks.

    test_shots : int, optional
        Number of test examples per class in each task. If `None`, then the
        number of test examples is equal to the number of training examples per
        class.

    seed : int, optional
        Random seed to be used in the meta-dataset.

    kwargs
        Additional arguments passed to the `TieredImagenet` class.

    See also
    --------
    `datasets.TieredImagenet` : Meta-dataset for the Tiered-Imagenet dataset.
    """
    defaults = {
        'transform': Compose([Resize(84), ToTensor()])
    }

    return helper_with_default(TieredImagenet, folder, shots, ways,
                               shuffle=shuffle, test_shots=test_shots,
                               seed=seed, defaults=defaults, **kwargs)

def cifar_fs(folder, shots, ways, shuffle=True, test_shots=None,
             seed=None, **kwargs):
    """Helper function to create a meta-dataset for the CIFAR-FS dataset.

    Parameters
    ----------
    folder : string
        Root directory where the dataset folder `cifar100` exists.

    shots : int
        Number of (training) examples per class in each task. This corresponds
        to `k` in `k-shot` classification.

    ways : int
        Number of classes per task. This corresponds to `N` in `N-way`
        classification.

    shuffle : bool (default: `True`)
        Shuffle the examples when creating the tasks.

    test_shots : int, optional
        Number of test examples per class in each task. If `None`, then the
        number of test examples is equal to the number of training examples per
        class.

    seed : int, optional
        Random seed to be used in the meta-dataset.

    kwargs
        Additional arguments passed to the `CIFARFS` class.

    See also
    --------
    `datasets.cifar100.CIFARFS` : Meta-dataset for the CIFAR-FS dataset.
    """
    return helper_with_default(CIFARFS, folder, shots, ways,
                               shuffle=shuffle, test_shots=test_shots,
                               seed=seed, defaults={}, **kwargs)

def fc100(folder, shots, ways, shuffle=True, test_shots=None,
          seed=None, **kwargs):
    """Helper function to create a meta-dataset for the FC100 dataset.

    Parameters
    ----------
    folder : string
        Root directory where the dataset folder `cifar100` exists.

    shots : int
        Number of (training) examples per class in each task. This corresponds
        to `k` in `k-shot` classification.

    ways : int
        Number of classes per task. This corresponds to `N` in `N-way`
        classification.

    shuffle : bool (default: `True`)
        Shuffle the examples when creating the tasks.

    test_shots : int, optional
        Number of test examples per class in each task. If `None`, then the
        number of test examples is equal to the number of training examples per
        class.

    seed : int, optional
        Random seed to be used in the meta-dataset.

    kwargs
        Additional arguments passed to the `FC100` class.

    See also
    --------
    `datasets.cifar100.FC100` : Meta-dataset for the FC100 dataset.
    """
    return helper_with_default(FC100, folder, shots, ways,
                               shuffle=shuffle, test_shots=test_shots,
                               seed=seed, defaults={}, **kwargs)

def cub(folder, shots, ways, shuffle=True, test_shots=None,
        seed=None, **kwargs):
    """Helper function to create a meta-dataset for the Caltech-UCSD Birds dataset.

    Parameters
    ----------
    folder : string
        Root directory where the dataset folder `cub` exists.

    shots : int
        Number of (training) examples per class in each task. This corresponds
        to `k` in `k-shot` classification.

    ways : int
        Number of classes per task. This corresponds to `N` in `N-way`
        classification.

    shuffle : bool (default: `True`)
        Shuffle the examples when creating the tasks.

    test_shots : int, optional
        Number of test examples per class in each task. If `None`, then the
        number of test examples is equal to the number of training examples per
        class.

    seed : int, optional
        Random seed to be used in the meta-dataset.

    kwargs
        Additional arguments passed to the `CUB` class.

    See also
    --------
    `datasets.cub.CUB` : Meta-dataset for the Caltech-UCSD Birds dataset.
    """
    image_size = 84
    defaults = {
        'transform': Compose([
                        Resize(int(image_size * 1.5)),
                        CenterCrop(image_size),
                        ToTensor()
                    ])
    }

    return helper_with_default(CUB, folder, shots, ways,
                               shuffle=shuffle, test_shots=test_shots,
                               seed=seed, defaults=defaults, **kwargs)

def doublemnist(folder, shots, ways, shuffle=True, test_shots=None,
                seed=None, **kwargs):
    """Helper function to create a meta-dataset for the Double MNIST dataset.

    Parameters
    ----------
    folder : string
        Root directory where the dataset folder `doublemnist` exists.

    shots : int
        Number of (training) examples per class in each task. This corresponds
        to `k` in `k-shot` classification.

    ways : int
        Number of classes per task. This corresponds to `N` in `N-way`
        classification.

    shuffle : bool (default: `True`)
        Shuffle the examples when creating the tasks.

    test_shots : int, optional
        Number of test examples per class in each task. If `None`, then the
        number of test examples is equal to the number of training examples per
        class.

    seed : int, optional
        Random seed to be used in the meta-dataset.

    kwargs
        Additional arguments passed to the `DoubleMNIST` class.

    See also
    --------
    `datasets.doublemnist.DoubleMNIST` : Meta-dataset for the Double MNIST dataset.
    """
    return helper_with_default(DoubleMNIST, folder, shots, ways,
                               shuffle=shuffle, test_shots=test_shots,
                               seed=seed, defaults={}, **kwargs)

def triplemnist(folder, shots, ways, shuffle=True, test_shots=None,
                seed=None, **kwargs):
    """Helper function to create a meta-dataset for the Triple MNIST dataset.

    Parameters
    ----------
    folder : string
        Root directory where the dataset folder `triplemnist` exists.

    shots : int
        Number of (training) examples per class in each task. This corresponds 
        to `k` in `k-shot` classification.

    ways : int
        Number of classes per task. This corresponds to `N` in `N-way` 
        classification.

    shuffle : bool (default: `True`)
        Shuffle the examples when creating the tasks.

    test_shots : int, optional
        Number of test examples per class in each task. If `None`, then the 
        number of test examples is equal to the number of training examples per 
        class.

    seed : int, optional
        Random seed to be used in the meta-dataset.

    kwargs
        Additional arguments passed to the `TripleMNIST` class.

    See also
    --------
    `datasets.triplemnist.TripleMNIST` : Meta-dataset for the Triple MNIST dataset.
    """
    return helper_with_default(TripleMNIST, folder, shots, ways,
                               shuffle=shuffle, test_shots=test_shots,
                               seed=seed, defaults={}, **kwargs)

def pascal5i(folder, shots, ways=1, shuffle=True, test_shots=None,
             seed=None, **kwargs):
    """Helper function to create a meta-dataset for the PASCAL-VOC dataset.

    Parameters
    ----------
    folder : string
        Root directory where the dataset folder `omniglot` exists.

    shots : int
        Number of (training) examples per class in each task. This corresponds
        to `k` in `k-shot` classification.

    ways : int
        Number of classes per task. This corresponds to `N` in `N-way`
        classification. Only supports 1-way currently

    shuffle : bool (default: `True`)
        Shuffle the examples when creating the tasks.

    test_shots : int, optional
        Number of test examples per class in each task. If `None`, then the
        number of test examples is equal to the number of training examples per
        class.

    seed : int, optional
        Random seed to be used in the meta-dataset.

    kwargs
        Additional arguments passed to the `Omniglot` class.

    """
    defaults = {
        'transform': SegmentationPairTransform(500),
        'class_augmentations': []
    }
    return helper_with_default(Pascal5i, folder, shots, ways,
                               shuffle=shuffle, test_shots=test_shots,
                               seed=seed, defaults=defaults, **kwargs)



import pytest

import numpy as np
import torch
import torch.nn.functional as F

from collections import OrderedDict
from torchmeta.modules import MetaLinear
from torchmeta.utils.r2d2 import ridge_regression


@pytest.mark.parametrize('reg_lambda', [0.1, 1.])
@pytest.mark.parametrize('use_woodbury', [None, True, False])
@pytest.mark.parametrize('scale', [True, False])
@pytest.mark.parametrize('bias', [True, False])
def test_ridge_regression(reg_lambda, use_woodbury, scale, bias):
    # Numpy
    num_classes = 3
    embeddings_np = np.random.randn(5, 7).astype(np.float32)
    targets_np = np.random.randint(0, num_classes, size=(5,))

    # PyTorch
    embeddings_th = torch.as_tensor(embeddings_np)
    targets_th = torch.as_tensor(targets_np)
    model = MetaLinear(7, 3, bias=bias)

    solution = ridge_regression(embeddings_th,
                                targets_th,
                                reg_lambda,
                                num_classes,
                                use_woodbury=use_woodbury,
                                scale=scale,
                                bias=bias)

    assert solution.weight.shape == (3, 7)
    if bias:
        assert solution.bias is not None
        assert solution.bias.shape == (3,)
    else:
        assert solution.bias is None

    # Optimality criterion
    # Check if the gradient of the L2-regularized MSE at the solution
    # is close to 0
    params = OrderedDict([('weight', solution.weight.requires_grad_())])
    if bias:
        params['bias'] = solution.bias.requires_grad_()

    logits = model(embeddings_th, params=params)
    targets_binary = F.one_hot(targets_th, num_classes=num_classes).float()

    # Least-square
    loss = F.mse_loss(logits, targets_binary, reduction='sum')
    if scale:
        loss /= embeddings_th.size(0)

    # L2-regularization
    loss += reg_lambda * torch.sum(solution.weight ** 2)
    if bias:
        loss += reg_lambda * torch.sum(solution.bias ** 2)
    loss.backward()

    np.testing.assert_allclose(solution.weight.grad.numpy(), 0., atol=1e-4)
    if bias:
        np.testing.assert_allclose(solution.bias.grad.numpy(), 0., atol=1e-4)


@pytest.mark.parametrize('reg_lambda', [0.1, 1.])
@pytest.mark.parametrize('use_woodbury', [None, True, False])
@pytest.mark.parametrize('scale', [True, False])
@pytest.mark.parametrize('bias', [True, False])
def test_ridge_regression_requires_grad(reg_lambda, use_woodbury, scale, bias):
    # Numpy
    num_classes = 3
    embeddings_np = np.random.randn(5, 7).astype(np.float32)
    targets_np = np.random.randint(0, num_classes, size=(5,))

    # PyTorch
    embeddings_th = torch.as_tensor(embeddings_np).requires_grad_()
    targets_th = torch.as_tensor(targets_np)
    model = MetaLinear(7, 3, bias=bias)

    solution = ridge_regression(embeddings_th,
                                targets_th,
                                reg_lambda,
                                num_classes,
                                use_woodbury=use_woodbury,
                                scale=scale,
                                bias=bias)
    params = OrderedDict([('weight', solution.weight)])
    if bias:
        params['bias'] = solution.bias

    # Compute loss on test/query samples
    test_embeddings = torch.randn(11, 7)
    test_logits = model(test_embeddings, params=params)
    test_targets = torch.randint(num_classes, size=(11,))
    loss = F.cross_entropy(test_logits, test_targets)

    # Backpropagation
    loss.backward()

    assert embeddings_th.grad is not None


@pytest.mark.parametrize('use_woodbury', [None, True, False])
@pytest.mark.parametrize('scale', [True, False])
@pytest.mark.parametrize('bias', [True, False])
def test_ridge_regression_regression_task(use_woodbury, scale, bias):
    # Numpy
    reg_lambda = 1.
    embeddings_np = np.random.randn(5, 7).astype(np.float32)
    targets_np = np.random.randn(5, 3).astype(np.float32)

    # PyTorch
    embeddings_th = torch.as_tensor(embeddings_np)
    targets_th = torch.as_tensor(targets_np)
    model = MetaLinear(7, 3, bias=bias)

    solution = ridge_regression(embeddings_th,
                                targets_th,
                                reg_lambda,
                                num_classes=None,
                                use_woodbury=use_woodbury,
                                scale=scale,
                                bias=bias)

    assert solution.weight.shape == (3, 7)
    if bias:
        assert solution.bias.shape == (3,)

    # Optimality criterion
    # Check if the gradient of the L2-regularized MSE at the solution
    # is close to 0
    params = OrderedDict([('weight', solution.weight.requires_grad_())])
    if bias:
        params['bias'] = solution.bias.requires_grad_()

    logits = model(embeddings_th, params=params)

    # Least-square
    loss = F.mse_loss(logits, targets_th, reduction='sum')
    if scale:
        loss /= embeddings_th.size(0)

    # L2-regularization
    loss += reg_lambda * torch.sum(solution.weight ** 2)
    if bias:
        loss += reg_lambda * torch.sum(solution.bias ** 2)
    loss.backward()

    np.testing.assert_allclose(solution.weight.grad.numpy(), 0., atol=1e-4)
    if bias:
        np.testing.assert_allclose(solution.bias.grad.numpy(), 0., atol=1e-4)



import pytest

import numpy as np
import torch
import torch.nn as nn
from collections import OrderedDict

from torchmeta.modules import MetaModule
from torchmeta.modules.sparse import MetaEmbedding, MetaEmbeddingBag


def test_metaembedding():
    meta_model = MetaEmbedding(5, 3, padding_idx=0)
    model = nn.Embedding(5, 3, padding_idx=0)

    assert isinstance(meta_model, MetaModule)
    assert isinstance(meta_model, nn.Embedding)

    # Set same weights for both models
    weight = torch.randn(5, 3)
    meta_model.weight.data.copy_(weight)
    model.weight.data.copy_(weight)

    inputs = torch.randint(5, size=(2, 7))

    outputs_torchmeta = meta_model(inputs, params=None)
    outputs_nn = model(inputs)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())


def test_metaembedding_params():
    meta_model = MetaEmbedding(5, 3, padding_idx=0)
    model = nn.Embedding(5, 3, padding_idx=0)

    params = OrderedDict()
    params['weight'] = torch.randn(5, 3)
    model.weight.data.copy_(params['weight'])

    inputs = torch.randint(5, size=(2, 7))

    outputs_torchmeta = meta_model(inputs, params=params)
    outputs_nn = model(inputs)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())


@pytest.mark.parametrize('mode', ['sum', 'mean', 'max'])
def test_metaembeddingbag(mode):
    meta_model = MetaEmbeddingBag(5, 3, mode=mode)
    model = nn.EmbeddingBag(5, 3, mode=mode)

    assert isinstance(meta_model, MetaModule)
    assert isinstance(meta_model, nn.EmbeddingBag)

    # Set same weights for both models
    weight = torch.randn(5, 3)
    meta_model.weight.data.copy_(weight)
    model.weight.data.copy_(weight)

    inputs = torch.randint(5, size=(2, 7))

    outputs_torchmeta = meta_model(inputs, params=None)
    outputs_nn = model(inputs)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())


@pytest.mark.parametrize('mode', ['sum', 'mean', 'max'])
def test_metaembeddingbag_params(mode):
    meta_model = MetaEmbeddingBag(5, 3, mode=mode)
    model = nn.EmbeddingBag(5, 3, mode=mode)

    params = OrderedDict()
    params['weight'] = torch.randn(5, 3)
    model.weight.data.copy_(params['weight'])

    inputs = torch.randint(5, size=(2, 7))

    outputs_torchmeta = meta_model(inputs, params=params)
    outputs_nn = model(inputs)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())



import pytest

import os
import torch

from torchmeta.utils.data import MetaDataset, Task
from torchmeta.datasets import helpers

is_local = (os.getenv('TORCHMETA_DATA_FOLDER') is not None)


@pytest.mark.skipif(not is_local, reason='Requires datasets downloaded locally')
@pytest.mark.parametrize('name', helpers.__all__)
@pytest.mark.parametrize('shots', [1, 5])
@pytest.mark.parametrize('split', ['train', 'val', 'test'])
def test_datasets_helpers(name, shots, split):
    function = getattr(helpers, name)
    folder = os.getenv('TORCHMETA_DATA_FOLDER')
    download = bool(os.getenv('TORCHMETA_DOWNLOAD', False))

    dataset = function(folder,
                       ways=5,
                       shots=shots,
                       test_shots=15,
                       meta_split=split,
                       download=download)

    assert isinstance(dataset, MetaDataset)

    task = dataset.sample_task()

    # Task is a dictionary with keys [train, test]
    assert isinstance(task, dict)
    assert set(task.keys()) == set(['train', 'test'])

    # Train
    assert isinstance(task['train'], Task)
    assert task['train'].num_classes == 5
    assert len(task['train']) == 5 * shots

    # Test
    assert isinstance(task['test'], Task)
    assert task['test'].num_classes == 5
    assert len(task['test']) == 5 * 15 # test_shots



import pytest

import torch
import torch.nn as nn

from collections import OrderedDict
from torchmeta.modules import MetaLinear, MetaModule, MetaSequential


class MetaModel(MetaModule):
    def __init__(self):
        super(MetaModel, self).__init__()
        self.features = MetaSequential(OrderedDict([
            ('linear1', nn.Linear(2, 3)), ('relu1', nn.ReLU()),
            ('linear2', nn.Linear(3, 5)), ('relu2', nn.ReLU())]))
        self.classifier = MetaLinear(5, 7, bias=False)

    def forward(self, inputs, params=None):
        features = self.features(inputs,
            params=self.get_subdict(params, 'features'))
        return self.classifier(features,
            params=self.get_subdict(params, 'classifier'))


def test_get_subdict_unknown_key():
    model = MetaModel()
    params = OrderedDict([('classifier.weight', torch.randn(7, 5))])

    inputs = torch.randn(11, 2)
    with pytest.warns(UserWarning):
        outputs = model(inputs, params=params)

    assert isinstance(outputs, torch.Tensor)
    assert outputs.shape == (11, 7)


def test_get_subdict_extra_key():
    model = MetaModel()
    params = OrderedDict([
        ('features.linear1.weight', torch.randn(3, 2)),
        ('classifier.weight', torch.randn(7, 5))
    ])

    inputs = torch.randn(11, 2)
    outputs = model(inputs, params=params)

    assert isinstance(outputs, torch.Tensor)
    assert outputs.shape == (11, 7)



import torch
import numpy as np

from collections import OrderedDict, defaultdict
from torchmeta.utils.data.task import Task, ConcatTask, SubsetTask
from torchmeta.transforms.utils import apply_wrapper

__all__ = ['Splitter', 'ClassSplitter', 'WeightedClassSplitter']


class Splitter(object):
    def __init__(self, splits, random_state_seed):
        self.splits = splits
        self.random_state_seed = random_state_seed
        self.seed(random_state_seed)

    def seed(self, seed):
        self.np_random = np.random.RandomState(seed=seed)

    def get_indices(self, task):
        if isinstance(task, ConcatTask):
            indices = self.get_indices_concattask(task)
        elif isinstance(task, Task):
            indices = self.get_indices_task(task)
        else:
            raise ValueError('The task must be of type `ConcatTask` or `Task`, '
                'Got type `{0}`.'.format(type(task)))
        return indices

    def get_indices_task(self, task):
        raise NotImplementedError('Method `get_indices_task` must be '
            'implemented in classes inherited from `Splitter`.')

    def get_indices_concattask(self, task):
        raise NotImplementedError('Method `get_indices_concattask` must be '
            'implemented in classes inherited from `Splitter`.')

    def _get_class_indices(self, task):
        class_indices = defaultdict(list)
        if task.num_classes is None: # Regression task
            class_indices['regression'] = range(len(task))
        else:
            for index in range(len(task)):
                sample = task[index]
                if (not isinstance(sample, tuple)) or (len(sample) < 2):
                    raise ValueError('In order to split the dataset in train/'
                        'test splits, `Splitter` must access the targets. Each '
                        'sample from a task must be a tuple with at least 2 '
                        'elements, with the last one being the target.')
                class_indices[sample[-1]].append(index)

            if len(class_indices) != task.num_classes:
                raise ValueError('The number of classes detected in `Splitter` '
                    '({0}) is different from the property `num_classes` ({1}) '
                    'in task `{2}`.'.format(len(class_indices),
                    task.num_classes, task))

        return class_indices

    def __call__(self, task):
        indices = self.get_indices(task)
        return OrderedDict([(split, SubsetTask(task, indices[split]))
            for split in self.splits])

    def __len__(self):
        return len(self.splits)


class ClassSplitter_(Splitter):
    def __init__(self, shuffle=True, num_samples_per_class=None,
                 num_train_per_class=None, num_test_per_class=None,
                 num_support_per_class=None, num_query_per_class=None,
                 random_state_seed=0):
        """
        Transforms a dataset into train/test splits for few-shot learning tasks,
        based on a fixed number of samples per class for each split. This is a
        dataset transformation to be applied as a `dataset_transform` in a
        `MetaDataset`.

        Parameters
        ----------
        shuffle : bool (default: `True`)
            Shuffle the data in the dataset before the split.

        num_samples_per_class : dict, optional
            Dictionary containing the names of the splits (as keys) and the
            corresponding number of samples per class in each split (as values).
            If not `None`, then the arguments `num_train_per_class`,
            `num_test_per_class`, `num_support_per_class` and
            `num_query_per_class` are ignored.

        num_train_per_class : int, optional
            Number of samples per class in the training split. This corresponds
            to the number of "shots" in "k-shot learning". If not `None`, this
            creates an item `train` for each task.

        num_test_per_class : int, optional
            Number of samples per class in the test split. If not `None`, this
            creates an item `test` for each task.

        num_support_per_class : int, optional
            Alias for `num_train_per_class`. If `num_train_per_class` is not
            `None`, then this argument is ignored. If not `None`, this creates
            an item `support` for each task.

        num_query_per_class : int, optional
            Alias for `num_test_per_class`. If `num_test_per_class` is not
            `None`, then this argument is ignored. If not `None`, this creates
            an item `query` for each task.

        random_state_seed : int, optional
            seed of the np.RandomState. Defaults to '0'.

        Examples
        --------
        >>> transform = ClassSplitter(num_samples_per_class={
        ...     'train': 5, 'test': 15})
        >>> dataset = Omniglot('data', num_classes_per_task=5,
        ...                    dataset_transform=transform, meta_train=True)
        >>> task = dataset.sample_task()
        >>> task.keys()
        ['train', 'test']
        >>> len(task['train']), len(task['test'])
        (25, 75)
        """
        self.shuffle = shuffle

        if num_samples_per_class is None:
            num_samples_per_class = OrderedDict()
            if num_train_per_class is not None:
                num_samples_per_class['train'] = num_train_per_class
            elif num_support_per_class is not None:
                num_samples_per_class['support'] = num_support_per_class
            if num_test_per_class is not None:
                num_samples_per_class['test'] = num_test_per_class
            elif num_query_per_class is not None:
                num_samples_per_class['query'] = num_query_per_class
        assert len(num_samples_per_class) > 0

        self._min_samples_per_class = sum(num_samples_per_class.values())
        super(ClassSplitter_, self).__init__(num_samples_per_class, random_state_seed)

    def get_indices_task(self, task):
        all_class_indices = self._get_class_indices(task)
        indices = OrderedDict([(split, []) for split in self.splits])

        for i, (name, class_indices) in enumerate(all_class_indices.items()):
            num_samples = len(class_indices)
            if num_samples < self._min_samples_per_class:
                raise ValueError('The number of samples for class `{0}` ({1}) '
                    'is smaller than the minimum number of samples per class '
                    'required by `ClassSplitter` ({2}).'.format(name,
                    num_samples, self._min_samples_per_class))

            if self.shuffle:
                seed = (hash(task) + i + self.random_state_seed) % (2 ** 32)
                dataset_indices = np.random.RandomState(seed).permutation(num_samples)
            else:
                dataset_indices = np.arange(num_samples)

            ptr = 0
            for split, num_split in self.splits.items():
                split_indices = dataset_indices[ptr:ptr + num_split]
                if self.shuffle:
                    self.np_random.shuffle(split_indices)
                indices[split].extend([class_indices[idx] for idx in split_indices])
                ptr += num_split

        return indices

    def get_indices_concattask(self, task):
        indices = OrderedDict([(split, []) for split in self.splits])
        cum_size = 0

        for dataset in task.datasets:
            num_samples = len(dataset)
            if num_samples < self._min_samples_per_class:
                raise ValueError('The number of samples for one class ({0}) '
                    'is smaller than the minimum number of samples per class '
                    'required by `ClassSplitter` ({1}).'.format(num_samples,
                    self._min_samples_per_class))

            if self.shuffle:
                seed = (hash(task) + hash(dataset) + self.random_state_seed) % (2 ** 32)
                dataset_indices = np.random.RandomState(seed).permutation(num_samples)
            else:
                dataset_indices = np.arange(num_samples)

            ptr = 0
            for split, num_split in self.splits.items():
                split_indices = dataset_indices[ptr:ptr + num_split]
                if self.shuffle:
                    self.np_random.shuffle(split_indices)
                indices[split].extend(split_indices + cum_size)
                ptr += num_split
            cum_size += num_samples

        return indices


class WeightedClassSplitter_(Splitter):
    def __init__(self, shuffle=True, min_num_samples=1, max_num_samples=None,
                 weights=None, train_weights=None, test_weights=None,
                 support_weights=None, query_weights=None,
                 force_equal_per_class=False, random_state_seed=0):
        """
        Transforms a dataset into train/test splits for few-shot learning tasks.
        The number of samples per class is proportional to the number of samples
        per class in the original dataset. This is a dataset transformation to
        be applied as a `dataset_transform` in a `MetaDataset`.

        Parameters
        ----------
        shuffle : bool (default: `True`)
            Shuffle the data in the dataset before the split.

        min_num_samples : int or dict, optional (default: 1)
            Minimum number of samples per class.

        max_num_samples : int or dict, optional
            Maximum number of samples per class.

        weights : dict, optional
            Dictionary containing the names of the splits (as keys) and the
            corresponding proportions of samples per class in each split (as
            values). If not `None`, then the arguments `train_weights`,
            `test_weights`, `support_weights` and `query_weights` are ignored.

        train_weights : float, optional
            Proportion of samples from each class in the training split. If not
            `None`, this creates an item `train` for each task.

        test_weights : float, optional
            Proportion of samples from each class in the training split. If not
            `None`, this creates an item `test` for each task.

        support_weights : float, optional
            Alias for `train_weights`. If `train_weights` is not `None`, then
            this argument is ignored. If not `None`, this creates an item
            `support` for each task.

        query_weights : float, optional
            Alias for `test_weights`. If `test_weights` is not `None`, then this
            argument is ignored. If not `None`, this creates an item `query` for
            each task.

        force_equal_per_class : bool (default: `False`)
            If `True`, then the number of samples per class is equal for each
            class; this is then proportional to the number of samples in the
            class with the minimum number of samples.

        random_state_seed : int, optional
            seed of the np.RandomState. Defaults to '0'.
        """
        self.shuffle = shuffle
        self.force_equal_per_class = force_equal_per_class

        if weights is None:
            weights = OrderedDict()
            if train_weights is not None:
                weights['train'] = train_weights
            elif support_weights is not None:
                weights['support'] = support_weights
            if test_weights is not None:
                weights['test'] = test_weights
            elif query_weights is not None:
                weights['query'] = query_weights
        assert len(weights) > 0
        assert sum(weights.values()) <= 1.

        if (min_num_samples is None) or isinstance(min_num_samples, int):
            if min_num_samples is None:
                min_num_samples = 0
            self.min_num_samples = OrderedDict([(split, min_num_samples)
                for split in weights])
        elif isinstance(min_num_samples, dict):
            self.min_num_samples = OrderedDict(min_num_samples)
        else:
            raise NotImplementedError('Argument `min_num_samples` in '
                '`WeightedClassSplitter` must be of type `dict` or `int`. Got '
                'type `{0}`.'.format(type(min_num_samples)))

        if max_num_samples is None:
            self.max_num_samples = None
        elif isinstance(max_num_samples, int):
            self.max_num_samples = OrderedDict([(split, max_num_samples)
                for split in weights])
        elif isinstance(max_num_samples, dict):
            self.max_num_samples = OrderedDict(max_num_samples)
        else:
            raise NotImplementedError('Argument `max_num_samples` in '
                '`WeightedClassSplitter` must be of type `dict` or `int`. Got '
                'type `{0}`.'.format(type(min_num_samples)))

        self._min_samples_per_class = sum(self.min_num_samples.values())
        super(WeightedClassSplitter_, self).__init__(weights, random_state_seed)

    def get_indices_task(self, task):
        all_class_indices = self._get_class_indices(task)
        indices = OrderedDict([(split, []) for split in self.splits])

        min_samples = min([len(class_indices) for class_indices
            in all_class_indices.values()])
        if min_samples < self._min_samples_per_class:
            raise ValueError('The smallest number of samples in a class ({0}) '
                    'is smaller than the minimum number of samples per class '
                    'required by `WeightedClassSplitter` ({1}).'.format(
                    min_samples, self._min_samples_per_class))

        for i, class_indices in enumerate(all_class_indices.values()):
            num_samples = (min_samples if self.force_equal_per_class
                else len(class_indices))
            if self.shuffle:
                seed = (hash(task) + i + self.random_state_seed) % (2 ** 32)
                dataset_indices = np.random.RandomState(seed).permutation(num_samples)
            else:
                dataset_indices = np.arange(num_samples)

            ptr = 0
            for split, weight in self.splits.items():
                num_split = max(self.min_num_samples[split], int(weight * num_samples))
                if self.max_num_samples is not None:
                    num_split = min(self.max_num_samples[split], num_split)
                split_indices = dataset_indices[ptr:ptr + num_split]
                if self.shuffle:
                    self.np_random.shuffle(split_indices)
                indices[split].extend([class_indices[idx] for idx in split_indices])
                ptr += num_split

        return indices

    def get_indices_concattask(self, task):
        indices = OrderedDict([(split, []) for split in self.splits])
        cum_size = 0

        min_samples = min([len(dataset) for dataset in task.datasets])
        if min_samples < self._min_samples_per_class:
            raise ValueError('The smallest number of samples in a class ({0}) '
                    'is smaller than the minimum number of samples per class '
                    'required by `WeightedClassSplitter` ({1}).'.format(
                    min_samples, self._min_samples_per_class))

        for dataset in task.datasets:
            num_samples = (min_samples if self.force_equal_per_class
                else len(dataset))
            if self.shuffle:
                seed = (hash(task) + hash(dataset) + self.random_state_seed) % (2 ** 32)
                dataset_indices = np.random.RandomState(seed).permutation(num_samples)
            else:
                dataset_indices = np.arange(num_samples)

            ptr = 0
            for split, weight in self.splits.items():
                num_split = max(self.min_num_samples, int(weight * num_samples))
                split_indices = dataset_indices[ptr:ptr + num_split]
                if self.shuffle:
                    self.np_random.shuffle(split_indices)
                indices[split].extend(split_indices + cum_size)
            cum_size += num_samples

        return indices


def ClassSplitter(task=None, *args, **kwargs):
    return apply_wrapper(ClassSplitter_(*args, **kwargs), task)

def WeightedClassSplitter(task=None, *args, **kwargs):
    return apply_wrapper(WeightedClassSplitter_(*args, **kwargs), task)



import sys
import numpy as np
import warnings
from copy import deepcopy

from itertools import combinations
from ordered_set import OrderedSet

from torchvision.transforms import Compose

from torchmeta.utils.data.task import ConcatTask
from torchmeta.transforms import FixedCategory, Categorical, DefaultTargetTransform
from torchmeta.transforms.utils import wrap_transform

__all__ = ['ClassDataset', 'MetaDataset', 'CombinationMetaDataset']


class ClassDataset(object):
    """Base class for a dataset of classes. Each item from a `ClassDataset` is 
    a dataset containing examples from the same class.

    Parameters
    ----------
    meta_train : bool (default: `False`)
        Use the meta-train split of the dataset. If set to `True`, then the
        arguments `meta_val` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_val : bool (default: `False`)
        Use the meta-validation split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_test : bool (default: `False`)
        Use the meta-test split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_val` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_split : string in {'train', 'val', 'test'}, optional
        Name of the split to use. This overrides the arguments `meta_train`, 
        `meta_val` and `meta_test`.

    class_augmentations : list of callable, optional
        A list of functions that augment the dataset with new classes. These classes 
        are transformations of existing classes. E.g. `transforms.HorizontalFlip()`.
    """
    def __init__(self, meta_train=False, meta_val=False, meta_test=False,
                 meta_split=None, class_augmentations=None):
        if meta_train + meta_val + meta_test == 0:
            if meta_split is None:
                raise ValueError('The meta-split is undefined. Use either the '
                    'argument `meta_train=True` (or `meta_val`/`meta_test`), or '
                    'the argument `meta_split="train"` (or "val"/"test").')
            elif meta_split not in ['train', 'val', 'test']:
                raise ValueError('Unknown meta-split name `{0}`. The meta-split '
                    'must be in [`train`, `val`, `test`].'.format(meta_split))
            meta_train = (meta_split == 'train')
            meta_val = (meta_split == 'val')
            meta_test = (meta_split == 'test')
        elif meta_train + meta_val + meta_test > 1:
            raise ValueError('Multiple arguments among `meta_train`, `meta_val` '
                'and `meta_test` are set to `True`. Exactly one must be set to '
                '`True`.')
        self.meta_train = meta_train
        self.meta_val = meta_val
        self.meta_test = meta_test
        self._meta_split = meta_split

        if class_augmentations is not None:
            if not isinstance(class_augmentations, list):
                raise TypeError('Unknown type for `class_augmentations`. '
                    'Expected `list`, got `{0}`.'.format(type(class_augmentations)))
            unique_augmentations = OrderedSet()
            for augmentations in class_augmentations:
                for transform in augmentations:
                    if transform in unique_augmentations:
                        warnings.warn('The class augmentation `{0}` already '
                            'exists in the list of class augmentations (`{1}`). '
                            'To avoid any duplicate, this transformation is '
                            'ignored.'.format(transform, repr(transform)),
                            UserWarning, stacklevel=2)
                    unique_augmentations.add(transform)
            class_augmentations = list(unique_augmentations)
        else:
            class_augmentations = []
        self.class_augmentations = class_augmentations

    def get_class_augmentation(self, index):
        transform_index = (index // self.num_classes) - 1
        if transform_index < 0:
            return None
        return self.class_augmentations[transform_index]

    def get_transform(self, index, transform=None):
        class_transform = self.get_class_augmentation(index)
        if class_transform is None:
            return transform
        if transform is None:
            return class_transform
        return Compose([class_transform, transform])

    def get_target_transform(self, index):
        class_transform = self.get_class_augmentation(index)
        return FixedCategory(class_transform)

    @property
    def meta_split(self):
        if self._meta_split is None:
            if self.meta_train:
                self._meta_split = 'train'
            elif self.meta_val:
                self._meta_split = 'val'
            elif self.meta_test:
                self._meta_split = 'test'
            else:
                raise NotImplementedError()
        return self._meta_split

    def __getitem__(self, index):
        raise NotImplementedError()

    @property
    def num_classes(self):
        raise NotImplementedError()

    def __len__(self):
        return self.num_classes * (len(self.class_augmentations) + 1)


class MetaDataset(object):
    """Base class for a meta-dataset.

    Parameters
    ----------
    meta_train : bool (default: `False`)
        Use the meta-train split of the dataset. If set to `True`, then the
        arguments `meta_val` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_val : bool (default: `False`)
        Use the meta-validation split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_test : bool (default: `False`)
        Use the meta-test split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_val` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_split : string in {'train', 'val', 'test'}, optional
        Name of the split to use. This overrides the arguments `meta_train`, 
        `meta_val` and `meta_test`.

    target_transform : callable, optional
        A function/transform that takes a target, and returns a transformed 
        version. See also `torchvision.transforms`.

    dataset_transform : callable, optional
        A function/transform that takes a dataset (ie. a task), and returns a 
        transformed version of it. E.g. `transforms.ClassSplitter()`.
    """
    def __init__(self, meta_train=False, meta_val=False, meta_test=False,
                 meta_split=None, target_transform=None, dataset_transform=None):
        if meta_train + meta_val + meta_test == 0:
            if meta_split is None:
                raise ValueError('The meta-split is undefined. Use either the '
                    'argument `meta_train=True` (or `meta_val`/`meta_test`), or '
                    'the argument `meta_split="train"` (or "val"/"test").')
            elif meta_split not in ['train', 'val', 'test']:
                raise ValueError('Unknown meta-split name `{0}`. The meta-split '
                    'must be in [`train`, `val`, `test`].'.format(meta_split))
            meta_train = (meta_split == 'train')
            meta_val = (meta_split == 'val')
            meta_test = (meta_split == 'test')
        elif meta_train + meta_val + meta_test > 1:
            raise ValueError('Multiple arguments among `meta_train`, `meta_val` '
                'and `meta_test` are set to `True`. Exactly one must be set to '
                '`True`.')
        self.meta_train = meta_train
        self.meta_val = meta_val
        self.meta_test = meta_test
        self._meta_split = meta_split
        self.target_transform = target_transform
        self.dataset_transform = dataset_transform
        self.seed()

    @property
    def meta_split(self):
        if self._meta_split is None:
            if self.meta_train:
                self._meta_split = 'train'
            elif self.meta_val:
                self._meta_split = 'val'
            elif self.meta_test:
                self._meta_split = 'test'
            else:
                raise NotImplementedError()
        return self._meta_split

    def seed(self, seed=None):
        self.np_random = np.random.RandomState(seed=seed)
        # Seed the dataset transform
        _seed_dataset_transform(self.dataset_transform, seed=seed)

    def __iter__(self):
        for index in range(len(self)):
            yield self[index]

    def sample_task(self):
        index = self.np_random.randint(len(self))
        return self[index]

    def __getitem__(self, index):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()


class CombinationMetaDataset(MetaDataset):
    """Base class for a meta-dataset, where the classification tasks are over 
    multiple classes from a `ClassDataset`.

    Parameters
    ----------
    dataset : `ClassDataset` instance
        A dataset of classes. Each item of `dataset` is a dataset, containing 
        all the examples from the same class.

    num_classes_per_task : int
        Number of classes per tasks. This corresponds to `N` in `N-way` 
        classification.

    target_transform : callable, optional
        A function/transform that takes a target, and returns a transformed 
        version. See also `torchvision.transforms`.

    dataset_transform : callable, optional
        A function/transform that takes a dataset (ie. a task), and returns a 
        transformed version of it. E.g. `transforms.ClassSplitter()`.
    """
    def __init__(self, dataset, num_classes_per_task, target_transform=None,
                 dataset_transform=None):
        if not isinstance(num_classes_per_task, int):
            raise TypeError('Unknown type for `num_classes_per_task`. Expected '
                '`int`, got `{0}`.'.format(type(num_classes_per_task)))
        self.dataset = dataset
        self.num_classes_per_task = num_classes_per_task
        # If no target_transform, then use a default target transform that
        # is well behaved for the `default_collate` function (assign class
        # augmentations ot integers).
        if target_transform is None:
            target_transform = DefaultTargetTransform(dataset.class_augmentations)

        super(CombinationMetaDataset, self).__init__(meta_train=dataset.meta_train,
            meta_val=dataset.meta_val, meta_test=dataset.meta_test,
            meta_split=dataset.meta_split, target_transform=target_transform,
            dataset_transform=dataset_transform)

    def __iter__(self):
        num_classes = len(self.dataset)
        for index in combinations(num_classes, self.num_classes_per_task):
            yield self[index]

    def sample_task(self):
        index = self.np_random.choice(len(self.dataset),
            size=self.num_classes_per_task, replace=False)
        return self[tuple(index)]

    def __getitem__(self, index):
        if isinstance(index, int):
            raise ValueError('The index of a `CombinationMetaDataset` must be '
                'a tuple of integers, and not an integer. For example, call '
                '`dataset[({0})]` to get a task with classes from 0 to {1} '
                '(got `{2}`).'.format(', '.join([str(idx)
                for idx in range(self.num_classes_per_task)]),
                self.num_classes_per_task - 1, index))
        assert len(index) == self.num_classes_per_task
        datasets = [self.dataset[i] for i in index]
        # Use deepcopy on `Categorical` target transforms, to avoid any side
        # effect across tasks.
        task = ConcatTask(datasets, self.num_classes_per_task,
            target_transform=wrap_transform(self.target_transform,
            self._copy_categorical, transform_type=Categorical))

        if self.dataset_transform is not None:
            task = self.dataset_transform(task)

        return task

    def _copy_categorical(self, transform):
        assert isinstance(transform, Categorical)
        transform.reset()
        if transform.num_classes is None:
            transform.num_classes = self.num_classes_per_task
        return deepcopy(transform)

    def __len__(self):
        num_classes, length = len(self.dataset), 1
        for i in range(1, self.num_classes_per_task + 1):
            length *= (num_classes - i + 1) / i

        if length > sys.maxsize:
            warnings.warn('The number of possible tasks in {0} is '
                'combinatorially large (equal to C({1}, {2})), and exceeds '
                'machine precision. Setting the length of the dataset to the '
                'maximum integer value, which undervalues the actual number of '
                'possible tasks in the dataset. Therefore the value returned by '
                '`len(dataset)` should not be trusted as being representative '
                'of the true number of tasks.'.format(self, len(self.dataset),
                self.num_classes_per_task), UserWarning, stacklevel=2)
            length = sys.maxsize
        return int(length)


def _seed_dataset_transform(transform, seed=None):
    if isinstance(transform, Compose):
        for subtransform in transform.transforms:
            _seed_dataset_transform(subtransform, seed=seed)
    elif hasattr(transform, 'seed'):
        transform.seed(seed=seed)



import torch.nn as nn
import torch.nn.functional as F

from collections import OrderedDict
from torch.nn.modules.utils import _single, _pair, _triple
from torchmeta.modules.module import MetaModule

class MetaConv1d(nn.Conv1d, MetaModule):
    __doc__ = nn.Conv1d.__doc__

    def forward(self, input, params=None):
        if params is None:
            params = OrderedDict(self.named_parameters())
        bias = params.get('bias', None)

        if self.padding_mode == 'circular':
            expanded_padding = ((self.padding[0] + 1) // 2, self.padding[0] // 2)
            return F.conv1d(F.pad(input, expanded_padding, mode='circular'),
                            params['weight'], bias, self.stride,
                            _single(0), self.dilation, self.groups)

        return F.conv1d(input, params['weight'], bias, self.stride,
                        self.padding, self.dilation, self.groups)

class MetaConv2d(nn.Conv2d, MetaModule):
    __doc__ = nn.Conv2d.__doc__

    def forward(self, input, params=None):
        if params is None:
            params = OrderedDict(self.named_parameters())
        bias = params.get('bias', None)

        if self.padding_mode == 'circular':
            expanded_padding = ((self.padding[1] + 1) // 2, self.padding[1] // 2,
                                (self.padding[0] + 1) // 2, self.padding[0] // 2)
            return F.conv2d(F.pad(input, expanded_padding, mode='circular'),
                            params['weight'], bias, self.stride,
                            _pair(0), self.dilation, self.groups)

        return F.conv2d(input, params['weight'], bias, self.stride,
                        self.padding, self.dilation, self.groups)

class MetaConv3d(nn.Conv3d, MetaModule):
    __doc__ = nn.Conv3d.__doc__

    def forward(self, input, params=None):
        if params is None:
            params = OrderedDict(self.named_parameters())
        bias = params.get('bias', None)

        if self.padding_mode == 'circular':
            expanded_padding = ((self.padding[2] + 1) // 2, self.padding[2] // 2,
                                (self.padding[1] + 1) // 2, self.padding[1] // 2,
                                (self.padding[0] + 1) // 2, self.padding[0] // 2)
            return F.conv3d(F.pad(input, expanded_padding, mode='circular'),
                            params['weight'], bias, self.stride,
                            _triple(0), self.dilation, self.groups)

        return F.conv3d(input, params['weight'], bias, self.stride,
                        self.padding, self.dilation, self.groups)



import os
import json
import h5py
import numpy as np
import torch
import copy

from ordered_set import OrderedSet
from torchmeta.utils.data import Task, MetaDataset
from torchmeta.datasets.utils import get_asset


class TCGA(MetaDataset):
    """
    The TCGA dataset [1]. A dataset of classification tasks over the values of
    an attribute, based on the gene expression data from patients diagnosed with
    specific types of cancer. This dataset is based on data from the Cancer
    Genome Atlas Program from the National Cancer Institute.

    Parameters
    ----------
    root : string
        Root directory where the dataset folder `omniglot` exists.

    meta_train : bool (default: `False`)
        Use the meta-train split of the dataset. If set to `True`, then the
        arguments `meta_val` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_val : bool (default: `False`)
        Use the meta-validation split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_test : bool (default: `False`)
        Use the meta-test split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_val` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_split : string in {'train', 'val', 'test'}, optional
        Name of the split to use. This overrides the arguments `meta_train`, 
        `meta_val` and `meta_test` if all three are set to `False`.

    min_samples_per_class : int (default: 5)
        Minimum number of samples per class in each classification task. This
        filters tasks for which the amount of data for one of the classes is
        too small.

    transform : callable, optional
        A function/transform that takes a `PIL` image, and returns a transformed 
        version. See also `torchvision.transforms`.

    target_transform : callable, optional
        A function/transform that takes a target, and returns a transformed 
        version. See also `torchvision.transforms`.

    dataset_transform : callable, optional
        A function/transform that takes a dataset (ie. a task), and returns a 
        transformed version of it. E.g. `transforms.ClassSplitter()`.

    download : bool (default: `False`)
        If `True`, downloads the files and processes the dataset in the root 
        directory (under the `tcga` folder). If the dataset is already 
        available, this does not download/process the dataset again.

    chunksize : int (default: 100)
        Size of the chunks to be processed when reading the CSV file. This is
        only used while downloading and converting the dataset to HDF5.

    preload : bool (default: `True`)
        Opens the gene expression dataset and keeps a reference to it in memory.
        This decreases the loading time of individual tasks.

    Notes
    -----
    A task is the combination of a cancer type and an attribute. The data is the
    gene expression of patients diagnosed with the cancer defined by the task.
    It consists in a vector of size `(20530,)`. The task is to classify the
    patients according to the attribute given by the task definition. The meta
    train/validation/test splits are over 137/29/29 tasks (ie. types of cancer).
    However, the number of tasks depends on the minimum number of samples per
    class specified by `min_samples_per_class`.

    References
    ----------
    .. [1] Samiei, M., Wurfl, T., Deleu, T., Weiss, M., Dutil, F., Fevens, T.,
           Boucher, G., Lemieux, S., and Cohen, J. P. (2019). The TCGA
           Meta-Dataset Clinical Benchmark. (https://arxiv.org/abs/1910.08636)
    """
    folder = 'tcga'
    clinical_matrix_url = 'https://tcga.xenahubs.net/download/TCGA.{0}.sampleMap/{0}_clinicalMatrix.gz'
    clinical_matrix_filename, _ = os.path.splitext(os.path.basename(clinical_matrix_url))
    gene_expression_filename = 'TCGA_HiSeqV2.hdf5'
    gene_expression_torrent = 'e4081b995625f9fc599ad860138acf7b6eb1cf6f'

    filename_tasks = '{0}_labels.json'

    _task_variables = None
    _cancers = None

    def __init__(self, root, meta_train=False, meta_val=False, meta_test=False, meta_split=None,
                 min_samples_per_class=5, transform=None, target_transform=None,
                 dataset_transform=None, download=False, chunksize=100, preload=True):
        super(TCGA, self).__init__(meta_train, meta_val, meta_test, meta_split,
            target_transform=target_transform, dataset_transform=dataset_transform)
        self.root = os.path.join(os.path.expanduser(root), self.folder)
        self.min_samples_per_class = min_samples_per_class
        self.transform = transform

        self._all_sample_ids = None
        self._gene_ids = None
        self._tasks = None

        if download:
            self.download(chunksize)

        self.preloaded = False
        self.gene_expression_data = None
        self.gene_expression_file = None
        if preload:
            self._preload_gene_expression_data()
            self.preloaded = True

        self.task_ids = self.get_task_ids()
        self.split_filename_tasks = os.path.join(self.root,
            self.filename_tasks.format(self.meta_split))

    def __len__(self):
        return len(self.task_ids)

    @property
    def gene_expression_path(self):
        filename = os.path.join(self.root, self.gene_expression_filename)
        if not os.path.isfile(filename):
            raise IOError('Gene expression data not found at {}'.format(filename))
        return filename

    @property
    def tasks(self):
        if self._tasks is None:
            with open(self.split_filename_tasks, 'r') as f:
                self._tasks = [task for task in json.load(f) if tuple(task) in self.task_ids]
        return self._tasks

    @property
    def cancers(self):
        if self._cancers is None:
            self._cancers = get_cancers()
        return self._cancers

    @property
    def task_variables(self):
        if self._task_variables is None:
            self._task_variables = frozenset(get_task_variables())
        return self._task_variables

    @property
    def gene_ids(self):
        if self._gene_ids is None:
            gene_ids_file = os.path.join(self.root, 'gene_ids.json')
            if not os.path.isfile(gene_ids_file):
                raise IOError('Gene id data not found at {}'.format(gene_ids_file))
            with open(gene_ids_file, 'r') as f:
                self._gene_ids = set(json.load(f))
        return self._gene_ids

    @property
    def all_sample_ids(self):
        if self._all_sample_ids is None:
            all_sample_ids_file = os.path.join(self.root, 'all_sample_ids.json')
            if not os.path.isfile(all_sample_ids_file):
                raise IOError('All sample id data not found at {}'.format(all_sample_ids_file))
            with open(all_sample_ids_file, 'r') as f:
                all_sample_ids = json.load(f)
            self._all_sample_ids = dict((v, k) for (k, v) in enumerate(all_sample_ids))
        return self._all_sample_ids

    def get_processed_filename(self, cancer):
        processed_folder = os.path.join(self.root, 'clinicalMatrices', 'processed')
        filename = '{0}.tsv'.format(self.clinical_matrix_filename.format(cancer))
        filepath = os.path.join(processed_folder, filename)
        if not os.path.isfile(filepath):
            raise IOError('Clinical matrix file not found at {}'.format(filepath))
        return filepath

    def __getitem__(self, index):
        import pandas as pd

        label, cancer = self.tasks[index]
        filename = self.get_processed_filename(cancer)
        dataframe = pd.read_csv(filename, sep='\t', index_col=0, header=0)
        labels = dataframe[label].dropna().astype('category')
        labels = labels[self.task_ids[(label, cancer)]]

        if self.gene_expression_file is not None:
            data = self.gene_expression_data[labels.index]
        else:
            with h5py.File(self.gene_expression_path, 'r') as f:
                data = f['expression_data'][labels.index]

        task = TCGATask((label, cancer), data, labels.cat.codes.tolist(),
                        labels.cat.categories.tolist(), transform=self.transform,
                        target_transform=self.target_transform)

        if self.dataset_transform is not None:
            task = self.dataset_transform(task)

        return task

    def _preload_gene_expression_data(self):
        self.gene_expression_file = h5py.File(self.gene_expression_path, 'r')
        self.gene_expression_data = self.gene_expression_file['expression_data']

    def _process_clinical_matrices(self):
        import pandas as pd
        clinical_matrices_folder = os.path.join(self.root, 'clinicalMatrices')
        processed_folder = os.path.join(clinical_matrices_folder, 'processed')
        if not os.path.exists(processed_folder):
            os.makedirs(processed_folder)

        col_in_task_variables = lambda col: (col == 'sampleID') or (col in self.task_variables)

        for cancer in self.cancers:
            filename = self.clinical_matrix_filename.format(cancer)
            filepath = os.path.join(clinical_matrices_folder, '{0}.tsv'.format(filename))
            processed = os.path.join(processed_folder, '{0}.tsv'.format(filename))

            if not os.path.isfile(processed):
                raw_df = pd.read_csv(filepath, sep='\t', index_col=0, header=0,
                                     usecols=col_in_task_variables)
                dataframe = raw_df[raw_df.index.isin(self.all_sample_ids)]
                dataframe.index = dataframe.index.map(lambda index: self.all_sample_ids[index])
                dataframe.index.names = ['index']
                dataframe = dataframe.sort_index(axis=0)
                dataframe.to_csv(processed, sep='\t')
        return True

    def get_task_ids(self):
        tasks = get_task_id_splits(self.meta_split)
        task_ids = dict()

        for task_id in tasks:
            indices, counts = tasks[task_id]
            enough_samples = all(count > self.min_samples_per_class for count in counts.values())
            if enough_samples:
                task_id = tuple(task_id.split('|', 1))
                task_ids[task_id] = indices

        return task_ids

    def download(self, chunksize=100):
        try:
            import gzip
            import shutil
            import pandas as pd
            from six.moves import urllib
            import academictorrents as at
        except ImportError as exception:
            raise ImportError('{0}. To use the TCGA dataset, you need to '
                'install the necessary dependencies with '
                '`pip install torchmeta[tcga]`.'.format(exception.message))

        clinical_matrices_folder = os.path.join(self.root, 'clinicalMatrices')
        if not os.path.exists(clinical_matrices_folder):
            os.makedirs(clinical_matrices_folder)

        for cancer in self.cancers:
            filename = self.clinical_matrix_filename.format(cancer)
            rawpath = os.path.join(clinical_matrices_folder, '{0}.gz'.format(filename))
            filepath = os.path.join(clinical_matrices_folder, '{0}.tsv'.format(filename))

            if os.path.isfile(filepath):
                continue

            if not os.path.exists(rawpath):
                print('Downloading `{0}.gz`...'.format(filename))
                url = self.clinical_matrix_url.format(cancer)
                urllib.request.urlretrieve(url, rawpath)

            print('Extracting `{0}.gz`...'.format(filename))
            with gzip.open(rawpath, 'rb') as gzf:
                with open(filepath, 'wb') as f:
                    shutil.copyfileobj(gzf, f)

        gene_expression_file = os.path.join(self.root, self.gene_expression_filename)
        if not os.path.isfile(gene_expression_file):
            from tqdm import tqdm
            print('Downloading `{0}` using `academictorrents`...'.format(
                self.gene_expression_filename))
            csv_file = at.get(self.gene_expression_torrent, datastore=self.root)
            print('Downloaded to: `{0}`'.format(csv_file))

            print('Converting TCGA CSV dataset to HDF5. This may take a while, '
                  'but only happens on the first run.')
            reader = pd.read_csv(csv_file, compression='gzip', sep='\t',
                                 header=0, index_col=0, chunksize=chunksize)
            shape = (10459, 20530)

            with tqdm(total=shape[1]) as pbar:
                with h5py.File(gene_expression_file, 'w') as f:
                    dataset = f.create_dataset('expression_data',
                                               shape=shape, dtype='f4')
                    gene_ids = []
                    for idx, chunk in enumerate(reader):
                        slice_ = slice(idx * chunksize, (idx + 1) * chunksize)
                        dataset[:, slice_] = chunk.T
                        gene_ids.extend(chunk.index)
                        pbar.update(chunk.shape[0])
                    all_sample_ids = chunk.columns.tolist()

            gene_ids_file = os.path.join(self.root, 'gene_ids.json')
            with open(gene_ids_file, 'w') as f:
                json.dump(gene_ids, f)

            all_sample_ids_file = os.path.join(self.root, 'all_sample_ids.json')
            with open(all_sample_ids_file, 'w') as f:
                json.dump(all_sample_ids, f)

            if os.path.isfile(csv_file):
                os.remove(csv_file)

            print('Done')

        self._process_clinical_matrices()

        # Create label files
        for split in ['train', 'val', 'test']:
            filename = os.path.join(self.root, self.filename_tasks.format(split))
            data = get_asset(self.folder, '{0}.json'.format(split), dtype='json')

            with open(filename, 'w') as f:
                labels = sorted([key.split('|', 1) for key in data])
                json.dump(labels, f)

        # Clean up
        for cancer in self.cancers:
            filename = self.clinical_matrix_filename.format(cancer)
            rawpath = os.path.join(clinical_matrices_folder, '{0}.gz'.format(filename))
            if os.path.isfile(rawpath):
                os.remove(rawpath)

    def close(self):
        if self.preloaded:
            self.gene_expression_file.close()
            self.gene_expression_data = None
            self.gene_expression_file = None
            self.preloaded = False

    def open(self):
        if self.preloaded:
            self._preload_gene_expression_data()
            self.preloaded = True


class TCGATask(Task):
    @classmethod
    def from_id(cls, root, task_id, transform=None, target_transform=None):
        import pandas as pd
        root = os.path.join(os.path.expanduser(root), TCGA.folder)
        gene_filepath = os.path.join(root, TCGA.gene_expression_filename)
        if not os.path.isfile(gene_filepath):
            raise IOError()

        label, cancer = task_id

        processed_folder = os.path.join(root, 'clinicalMatrices', 'processed')
        filename = '{0}.tsv'.format(TCGA.clinical_matrix_filename.format(cancer))
        filepath = os.path.join(processed_folder, filename)
        if not os.path.isfile(filepath):
            raise IOError()

        dataframe = pd.read_csv(filepath, sep='\t', index_col=0, header=0)
        labels = dataframe[label].dropna().astype('category')

        with h5py.File(gene_filepath, 'r') as f:
            data = f['expression_data'][labels.index]

        return cls(task_id, data, labels.cat.codes.tolist(),
                   labels.cat.categories.tolist(), transform=transform,
                   target_transform=target_transform)

    def __init__(self, task_id, data, labels, categories,
                 transform=None, target_transform=None):
        super(TCGATask, self).__init__(task_id, len(categories),
            transform=transform, target_transform=target_transform)
        self.id = task_id
        self.data = data
        self.labels = labels
        self.categories = categories

    @property
    def input_size(self):
        return len(self.data[0])

    def __len__(self):
        return len(self.labels)

    def __iter__(self):
        for index in range(len(self)):
            yield self[index]

    def __getitem__(self, index):
        sample = self.data[index]
        target = self.labels[index]

        if self.transform is not None:
            sample = self.transform(sample)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return (sample, target)


def _assign_samples(tcga_metadataset):
    import pandas as pd
    import munkres

    blacklist = []
    sample_to_task_assignment = {}
    for cancer in get_cancers():
        filename = tcga_metadataset.get_processed_filename(cancer)
        dataframe = pd.read_csv(filename, sep='\t', index_col=0, header=0)
        dataframe = dataframe.drop(blacklist, errors='ignore')
        permutation = dataframe.index[torch.randperm(len(dataframe.index))]
        dataframe = dataframe.reindex(permutation)
        labels = dataframe.notna()
        labels = labels.applymap(lambda x: 1. if x else munkres.DISALLOWED)
        all_disallowed = labels.apply(lambda x: True if all(x == munkres.DISALLOWED) else False, axis=1)
        labels = labels.drop(labels[all_disallowed].index)

        matrix = labels.values
        shape = matrix.shape
        # The +5 allows for some slack in the assignment
        # which is necessary for the used implementation to converge on BRCA
        repeats = np.int(np.ceil(shape[0] / shape[1])) + 5
        expanded_matrix = np.tile(matrix, (1, repeats))

        indices = munkres.Munkres().compute(expanded_matrix)
        mapped_indices = [(a, b % shape[1]) for a, b in indices]

        for index, mapped_index in mapped_indices:
            sample_to_task_assignment.setdefault((dataframe.columns[mapped_index], cancer), []).append(
                dataframe.index[index])

        blacklist.extend(dataframe.index.tolist())

    return sample_to_task_assignment


def _expand_sample_usage(meta_dataset, all_allowed_samples, additional_samples):
    expanded_metadataset = {}
    all_samples_of_metadataset = OrderedSet()
    for key, value in meta_dataset.items():
        all_samples_of_metadataset.update(value)
    all_samples_of_metadataset.update(additional_samples)

    used_additional_samples = OrderedSet()
    for key in meta_dataset.keys():
        allowed_samples = set(all_allowed_samples[key])
        intersection = allowed_samples.intersection(all_samples_of_metadataset)
        expanded_metadataset[key] = list(intersection)
        used_additional_samples = additional_samples.intersection(intersection)

    return expanded_metadataset, used_additional_samples


def _split_tcga(tcga_metadataset, counts):
    all_allowed_samples = tcga_metadataset.task_ids

    # We first uniquely assing every sample to a task
    sample_to_task_assignment = _assign_samples(tcga_metadataset)

    keys = [i for i in all_allowed_samples.keys()]
    difference = set(sample_to_task_assignment.keys()).difference(set(keys))

    unassigned_samples = OrderedSet()
    for key in difference:
        unassigned_samples.update(sample_to_task_assignment[key])

    # Second we split the metadataset
    # with a torch-based random sample
    permutation = torch.randperm(len(keys)).numpy()

    metadatasets = []
    start = 0
    end = 0
    for count in counts:
        end += count
        current_keys = [keys[index] for index in permutation[start:end]]
        metadatasets.append({key: sample_to_task_assignment[key] for key in current_keys})
        start = end

    expanded_metadatasets = [None] * len(metadatasets)
    order = np.argsort([len(metadataset) for metadataset in metadatasets])

    # Finally we expand the tasks by reusing samples wherever possible in the sets
    blacklist = OrderedSet()
    for i in order:
        additional_samples = unassigned_samples.difference(blacklist)
        expanded_metadataset, used_additional_samples = _expand_sample_usage(
            metadatasets[i], all_allowed_samples, additional_samples)
        expanded_metadatasets[i] = (expanded_metadataset)
        blacklist.update(used_additional_samples)

    tcga_metadatasets = []
    tcga_metadataset.close()
    preloaded = tcga_metadataset.preloaded
    for metadataset in expanded_metadatasets:
        current_tcga_metadataset = copy.deepcopy(tcga_metadataset)
        current_tcga_metadataset.task_ids = metadataset
        if preloaded:
            current_tcga_metadataset.open()
        tcga_metadatasets.append(current_tcga_metadataset)

    return tcga_metadatasets


def get_cancers():
    return get_asset(TCGA.folder, 'cancers.json', dtype='json')


def get_task_variables():
    return get_asset(TCGA.folder, 'task_variables.json', dtype='json')


def get_task_id_splits(meta_split):
    return get_asset(TCGA.folder, '{}.json'.format(meta_split), dtype='json')



from setuptools import setup, find_packages
from os import path
import sys

from io import open

extras = {
    'tcga': ['pandas~=0.24.0', 'academictorrents~=2.1.0', 'six~=1.11.0'],
}

here = path.abspath(path.dirname(__file__))

sys.path.insert(0, path.join(here, 'torchmeta'))
from version import VERSION

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='torchmeta',
    version=VERSION,
    description='Dataloaders for meta-learning in Pytorch',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    author='Tristan Deleu',
    author_email='tristan.deleu@gmail.com',
    url='https://github.com/tristandeleu/pytorch-meta',
    keywords=['meta-learning', 'pytorch', 'few-shot', 'few-shot learning'],
    packages=find_packages(exclude=['data', 'contrib', 'docs', 'tests', 'examples']),
    install_requires=[
        'torch>=1.4.0,<1.8.0',
        'torchvision>=0.5.0,<0.9.0',
        'numpy>=1.14.0',
        'Pillow>=7.0.0',
        'h5py',
        'tqdm>=4.0.0',
        'requests',  # Required by Torchvision
        'ordered-set'
    ],
    extras_require=extras,
    package_data={'torchmeta': ['torchmeta/datasets/assets/*']},
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: MIT License',
    ],
)



from torchmeta import datasets
from torchmeta import modules
from torchmeta import toy
from torchmeta import transforms
from torchmeta import utils

from torchmeta.version import VERSION as __version__



import numpy as np
import io

from PIL import Image
from torch.utils.data import Dataset

from torchmeta.utils.data.dataset import CombinationMetaDataset


class NonEpisodicWrapper(Dataset):
    """Non-episodic wrapper to convert a CombinationMetaDataset into a standard
    PyTorch Dataset, compatible with (non-episodic) training.

    Parameters
    ----------
    dataset : `CombinationMetaDataset` instance
        The meta-dataset to be wrapped around.

    target_transform : callable, optional
        A function/transform that takes a target, and returns a transformed 
        version. See also `torchvision.transforms`.
    """
    def __init__(self, dataset, target_transform=None):
        super(NonEpisodicWrapper, self).__init__()
        if not isinstance(dataset, CombinationMetaDataset):
            raise ValueError('`NonEpisodicWrapper` can only be wrapped around a '
                '`CombinationMetaDataset`. The dataset `{0}` is not an instance '
                'of `CombinationMetaDataset`.'.format(dataset))

        self.dataset = dataset
        self.target_transform = target_transform

        class_dataset = self.dataset.dataset
        self._labels, pointer = [], 0
        self._offsets = np.zeros((class_dataset.num_classes,), dtype=np.int_)
        for index, label in enumerate(class_dataset.labels):
            if isinstance(label, list):
                label = '/'.join(label)
            num_samples = len(class_dataset.data[label])
            self._labels.append(label)
            self._offsets[index] = pointer
            pointer += num_samples
        self._num_samples = pointer

    def __getitem__(self, index):
        class_dataset = self.dataset.dataset

        class_index = np.maximum(np.searchsorted(self._offsets,
            index % self.num_samples, side='left') - 1, 0)
        offset = (index % self.num_samples) - self._offsets[class_index]
        label = self._labels[class_index]

        array = class_dataset.data[label][offset]
        image = (Image.open(io.BytesIO(array))
            if array.ndim < 2 else Image.fromarray(array))

        class_augmented_index = (class_dataset.num_classes
            * (index // self.num_samples) + class_index)
        transform = class_dataset.get_transform(class_augmented_index,
                                                class_dataset.transform)
        if transform is not None:
            image = transform(image)

        class_transform = class_dataset.get_class_augmentation(class_augmented_index)
        label = (label, index // self.num_samples)
        if self.target_transform is not None:
            label = self.target_transform(label)

        return image, label

    @property
    def num_samples(self):
        return self._num_samples

    @property
    def num_classes(self):
        num_augmentations = len(self.dataset.dataset.class_augmentations)
        return len(self._labels) * (num_augmentations + 1)

    def __len__(self):
        num_augmentations = len(self.dataset.dataset.class_augmentations)
        return self.num_samples * (num_augmentations + 1)



import pytest

import numpy as np
import torch
import torch.nn as nn
from collections import OrderedDict

from torchmeta.modules import MetaModule
from torchmeta.modules.activation import MetaMultiheadAttention


@pytest.mark.parametrize('bias', [True, False])
@pytest.mark.parametrize('add_bias_kv', [True, False])
@pytest.mark.parametrize('kdim,vdim', [(None, None), (7, 11)])
def test_metamultiheadattention(bias, add_bias_kv, kdim, vdim):
    meta_model = MetaMultiheadAttention(3 * 5, 3,
                                        bias=bias,
                                        add_bias_kv=add_bias_kv,
                                        kdim=kdim, vdim=vdim)
    model = nn.MultiheadAttention(3 * 5, 3,
                                  bias=bias,
                                  add_bias_kv=add_bias_kv,
                                  kdim=kdim, vdim=vdim)

    assert isinstance(meta_model, MetaModule)
    assert isinstance(meta_model, nn.MultiheadAttention)

    # Set same weights for both models
    if not meta_model._qkv_same_embed_dim:
        q_proj_weight = torch.randn(3 * 5, 3 * 5)
        meta_model.q_proj_weight.data.copy_(q_proj_weight)
        model.q_proj_weight.data.copy_(q_proj_weight)

        k_proj_weight = torch.randn(3 * 5, meta_model.kdim)
        meta_model.k_proj_weight.data.copy_(k_proj_weight)
        model.k_proj_weight.data.copy_(k_proj_weight)

        v_proj_weight = torch.randn(3 * 5, meta_model.vdim)
        meta_model.v_proj_weight.data.copy_(v_proj_weight)
        model.v_proj_weight.data.copy_(v_proj_weight)

    else:
        in_proj_weight = torch.randn(3 * 3 * 5, 3 * 5)
        meta_model.in_proj_weight.data.copy_(in_proj_weight)
        model.in_proj_weight.data.copy_(in_proj_weight)

    if bias:
        in_proj_bias = torch.randn(3 * 3 * 5)
        meta_model.in_proj_bias.data.copy_(in_proj_bias)
        model.in_proj_bias.data.copy_(in_proj_bias)

    if add_bias_kv:
        bias_k = torch.randn(1, 1, 3 * 5)
        meta_model.bias_k.data.copy_(bias_k)
        model.bias_k.data.copy_(bias_k)

        bias_v = torch.randn(1, 1, 3 * 5)
        meta_model.bias_v.data.copy_(bias_v)
        model.bias_v.data.copy_(bias_v)

    out_proj_weight = torch.randn(3 * 5, 3 * 5)
    meta_model.out_proj.weight.data.copy_(out_proj_weight)
    model.out_proj.weight.data.copy_(out_proj_weight)

    out_proj_bias = torch.randn(3 * 5)
    meta_model.out_proj.bias.data.copy_(out_proj_bias)
    model.out_proj.bias.data.copy_(out_proj_bias)

    query = torch.randn(13, 17, 3 * 5)
    key = torch.randn(19, 17, 3 * 5 if (kdim is None) else kdim)
    value = torch.randn(19, 17, 3 * 5 if (vdim is None) else vdim)

    outputs_torchmeta, weights_torchmeta = meta_model(query, key, value, params=None)
    outputs_nn, weights_nn = model(query, key, value)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())
    np.testing.assert_equal(weights_torchmeta.detach().numpy(),
                            weights_nn.detach().numpy())


@pytest.mark.parametrize('bias', [True, False])
@pytest.mark.parametrize('add_bias_kv', [True, False])
@pytest.mark.parametrize('kdim,vdim', [(None, None), (7, 11)])
def test_metamultiheadattention_params(bias, add_bias_kv, kdim, vdim):
    meta_model = MetaMultiheadAttention(3 * 5, 3,
                                        bias=bias,
                                        add_bias_kv=add_bias_kv,
                                        kdim=kdim, vdim=vdim)
    model = nn.MultiheadAttention(3 * 5, 3,
                                  bias=bias,
                                  add_bias_kv=add_bias_kv,
                                  kdim=kdim, vdim=vdim)

    params = OrderedDict()
    if not meta_model._qkv_same_embed_dim:
        params['q_proj_weight'] = torch.randn(3 * 5, 3 * 5)
        model.q_proj_weight.data.copy_(params['q_proj_weight'])

        params['k_proj_weight'] = torch.randn(3 * 5, meta_model.kdim)
        model.k_proj_weight.data.copy_(params['k_proj_weight'])

        params['v_proj_weight'] = torch.randn(3 * 5, meta_model.vdim)
        model.v_proj_weight.data.copy_(params['v_proj_weight'])

    else:
        params['in_proj_weight'] = torch.randn(3 * 3 * 5, 3 * 5)
        model.in_proj_weight.data.copy_(params['in_proj_weight'])

    if bias:
        params['in_proj_bias'] = torch.randn(3 * 3 * 5)
        model.in_proj_bias.data.copy_(params['in_proj_bias'])

    if add_bias_kv:
        params['bias_k'] = torch.randn(1, 1, 3 * 5)
        model.bias_k.data.copy_(params['bias_k'])

        params['bias_v'] = torch.randn(1, 1, 3 * 5)
        model.bias_v.data.copy_(params['bias_v'])

    params['out_proj.weight'] = torch.randn(3 * 5, 3 * 5)
    model.out_proj.weight.data.copy_(params['out_proj.weight'])

    params['out_proj.bias'] = torch.randn(3 * 5)
    model.out_proj.bias.data.copy_(params['out_proj.bias'])

    query = torch.randn(13, 17, 3 * 5)
    key = torch.randn(19, 17, 3 * 5 if (kdim is None) else kdim)
    value = torch.randn(19, 17, 3 * 5 if (vdim is None) else vdim)

    outputs_torchmeta, weights_torchmeta = meta_model(query, key, value, params=params)
    outputs_nn, weights_nn = model(query, key, value)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())
    np.testing.assert_equal(weights_torchmeta.detach().numpy(),
                            weights_nn.detach().numpy())


@pytest.mark.parametrize('bias', [True, False])
@pytest.mark.parametrize('add_bias_kv', [True, False])
@pytest.mark.parametrize('kdim,vdim', [(None, None), (7, 11)])
def test_metamultiheadattention_meta_named_parameters(bias, add_bias_kv, kdim, vdim):
    meta_model = MetaMultiheadAttention(3 * 5, 3,
                                        bias=bias,
                                        add_bias_kv=add_bias_kv,
                                        kdim=kdim, vdim=vdim)
    params = OrderedDict(meta_model.meta_named_parameters())
    param_names = set(params)

    if not meta_model._qkv_same_embed_dim:
        assert {'q_proj_weight', 'k_proj_weight', 'v_proj_weight'} <= param_names
    else:
        assert 'in_proj_weight' in param_names

    if bias:
        'in_proj_bias' in param_names

    if add_bias_kv:
        assert {'bias_k', 'bias_v'} <= param_names

    assert {'out_proj.weight', 'out_proj.bias'} <= param_names



import torch
import warnings

from torch.nn import DataParallel as DataParallel_
from torchmeta.modules.module import MetaModule
from collections import OrderedDict

from torch.nn.parallel import parallel_apply
from torch.nn.parallel.scatter_gather import scatter_kwargs
from torch.nn.parallel.replicate import _broadcast_coalesced_reshape


class DataParallel(DataParallel_, MetaModule):
    __doc__ = DataParallel_.__doc__

    def scatter(self, inputs, kwargs, device_ids):
        if not isinstance(self.module, MetaModule):
            return super(DataParallel, self).scatter(inputs, kwargs, device_ids)

        params = kwargs.pop('params', None)
        inputs_, kwargs_ = scatter_kwargs(inputs, kwargs, device_ids, dim=self.dim)
        # Add params argument unchanged back in kwargs
        replicas = self._replicate_params(params, inputs_, device_ids,
                                          detach=not torch.is_grad_enabled())
        kwargs_ = tuple(dict(params=replica, **kwarg)
                        for (kwarg, replica) in zip(kwargs_, replicas))
        return inputs_, kwargs_

    def _replicate_params(self, params, inputs, device_ids, detach=False):
        if params is None:
            module_params = OrderedDict(self.module.named_parameters())
        else:
            # Temporarily disable the warning if no parameter with key prefix
            # `module` was found. In that case, the original params dictionary
            # is used.
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                module_params = self.get_subdict(params, key='module')
            if module_params is None:
                module_params = params

        replicas = _broadcast_coalesced_reshape(list(module_params.values()),
                                                device_ids[:len(inputs)],
                                                detach)
        replicas = tuple(OrderedDict(zip(module_params.keys(), replica))
                         for replica in replicas)
        return replicas



import warnings

from torchmeta.toy import Sinusoid, Harmonic
from torchmeta.transforms import ClassSplitter

def sinusoid(shots, shuffle=True, test_shots=None, seed=None, **kwargs):
    """Helper function to create a meta-dataset for the Sinusoid toy dataset.

    Parameters
    ----------
    shots : int
        Number of (training) examples in each task. This corresponds to `k` in
        `k-shot` classification.

    shuffle : bool (default: `True`)
        Shuffle the examples when creating the tasks.

    test_shots : int, optional
        Number of test examples in each task. If `None`, then the number of test
        examples is equal to the number of training examples in each task.

    seed : int, optional
        Random seed to be used in the meta-dataset.

    kwargs
        Additional arguments passed to the `Sinusoid` class.

    See also
    --------
    `torchmeta.toy.Sinusoid` : Meta-dataset for the Sinusoid toy dataset.
    """
    if 'num_samples_per_task' in kwargs:
        warnings.warn('Both arguments `shots` and `num_samples_per_task` were '
            'set in the helper function for the number of samples in each task. '
            'Ignoring the argument `shots`.', stacklevel=2)
        if test_shots is not None:
            shots = kwargs['num_samples_per_task'] - test_shots
            if shots <= 0:
                raise ValueError('The argument `test_shots` ({0}) is greater '
                    'than the number of samples per task ({1}). Either use the '
                    'argument `shots` instead of `num_samples_per_task`, or '
                    'increase the value of `num_samples_per_task`.'.format(
                    test_shots, kwargs['num_samples_per_task']))
        else:
            shots = kwargs['num_samples_per_task'] // 2
    if test_shots is None:
        test_shots = shots

    dataset = Sinusoid(num_samples_per_task=shots + test_shots, **kwargs)
    dataset = ClassSplitter(dataset, shuffle=shuffle,
        num_train_per_class=shots, num_test_per_class=test_shots)
    dataset.seed(seed)

    return dataset

def harmonic(shots, shuffle=True, test_shots=None, seed=None, **kwargs):
    """Helper function to create a meta-dataset for the Harmonic toy dataset.

    Parameters
    ----------
    shots : int
        Number of (training) examples in each task. This corresponds to `k` in
        `k-shot` classification.

    shuffle : bool (default: `True`)
        Shuffle the examples when creating the tasks.

    test_shots : int, optional
        Number of test examples in each task. If `None`, then the number of test
        examples is equal to the number of training examples in each task.

    seed : int, optional
        Random seed to be used in the meta-dataset.

    kwargs
        Additional arguments passed to the `Harmonic` class.

    See also
    --------
    `torchmeta.toy.Harmonic` : Meta-dataset for the Harmonic toy dataset.
    """
    if 'num_samples_per_task' in kwargs:
        warnings.warn('Both arguments `shots` and `num_samples_per_task` were '
            'set in the helper function for the number of samples in each task. '
            'Ignoring the argument `shots`.', stacklevel=2)
        if test_shots is not None:
            shots = kwargs['num_samples_per_task'] - test_shots
            if shots <= 0:
                raise ValueError('The argument `test_shots` ({0}) is greater '
                    'than the number of samples per task ({1}). Either use the '
                    'argument `shots` instead of `num_samples_per_task`, or '
                    'increase the value of `num_samples_per_task`.'.format(
                    test_shots, kwargs['num_samples_per_task']))
        else:
            shots = kwargs['num_samples_per_task'] // 2
    if test_shots is None:
        test_shots = shots

    dataset = Harmonic(num_samples_per_task=shots + test_shots, **kwargs)
    dataset = ClassSplitter(dataset, shuffle=shuffle,
        num_train_per_class=shots, num_test_per_class=test_shots)
    dataset.seed(seed)

    return dataset



import torch
import torch.nn.functional as F

__all__ = ['get_num_samples', 'get_prototypes', 'prototypical_loss']


def get_num_samples(targets, num_classes, dtype=None):
    batch_size = targets.size(0)
    with torch.no_grad():
        ones = torch.ones_like(targets, dtype=dtype)
        num_samples = ones.new_zeros((batch_size, num_classes))
        num_samples.scatter_add_(1, targets, ones)
    return num_samples


def get_prototypes(embeddings, targets, num_classes):
    """Compute the prototypes (the mean vector of the embedded training/support 
    points belonging to its class) for each classes in the task.

    Parameters
    ----------
    embeddings : `torch.FloatTensor` instance
        A tensor containing the embeddings of the support points. This tensor 
        has shape `(batch_size, num_examples, embedding_size)`.

    targets : `torch.LongTensor` instance
        A tensor containing the targets of the support points. This tensor has 
        shape `(batch_size, num_examples)`.

    num_classes : int
        Number of classes in the task.

    Returns
    -------
    prototypes : `torch.FloatTensor` instance
        A tensor containing the prototypes for each class. This tensor has shape
        `(batch_size, num_classes, embedding_size)`.
    """
    batch_size, embedding_size = embeddings.size(0), embeddings.size(-1)
    
    num_samples = get_num_samples(targets, num_classes, dtype=embeddings.dtype)
    num_samples.unsqueeze_(-1)
    num_samples = torch.max(num_samples, torch.ones_like(num_samples))

    prototypes = embeddings.new_zeros((batch_size, num_classes, embedding_size))
    indices = targets.unsqueeze(-1).expand_as(embeddings)
    prototypes.scatter_add_(1, indices, embeddings).div_(num_samples)

    return prototypes


def prototypical_loss(prototypes, embeddings, targets, **kwargs):
    """Compute the loss (i.e. negative log-likelihood) for the prototypical 
    network, on the test/query points.

    Parameters
    ----------
    prototypes : `torch.FloatTensor` instance
        A tensor containing the prototypes for each class. This tensor has shape 
        `(batch_size, num_classes, embedding_size)`.

    embeddings : `torch.FloatTensor` instance
        A tensor containing the embeddings of the query points. This tensor has 
        shape `(batch_size, num_examples, embedding_size)`.

    targets : `torch.LongTensor` instance
        A tensor containing the targets of the query points. This tensor has 
        shape `(batch_size, num_examples)`.

    Returns
    -------
    loss : `torch.FloatTensor` instance
        The negative log-likelihood on the query points.
    """
    squared_distances = torch.sum((prototypes.unsqueeze(2)
        - embeddings.unsqueeze(1)) ** 2, dim=-1)
    return F.cross_entropy(-squared_distances, targets, **kwargs)



from torchmeta.utils import data
from torchmeta.utils.gradient_based import gradient_update_parameters
from torchmeta.utils.metrics import hardness_metric
from torchmeta.utils.prototype import get_num_samples, get_prototypes, prototypical_loss
from torchmeta.utils.matching import pairwise_cosine_similarity, matching_log_probas, matching_probas, matching_loss
from torchmeta.utils.r2d2 import ridge_regression

__all__ = [
    'data',
    'gradient_update_parameters',
    'hardness_metric',
    'get_num_samples',
    'get_prototypes',
    'prototypical_loss',
    'pairwise_cosine_similarity',
    'matching_log_probas',
    'matching_probas',
    'matching_loss',
    'ridge_regression'
]



import torch.nn as nn
import torch.nn.functional as F

from collections import OrderedDict
from torchmeta.modules.module import MetaModule
from torchmeta.modules.linear import MetaLinear

class MetaMultiheadAttention(nn.MultiheadAttention, MetaModule):
    __doc__ = nn.MultiheadAttention.__doc__

    def __init__(self, *args, **kwargs):
        super(MetaMultiheadAttention, self).__init__(*args, **kwargs)
        self.out_proj = MetaLinear(self.embed_dim, self.embed_dim, bias=True)

    def forward(self, query, key, value, key_padding_mask=None,
                need_weights=True, attn_mask=None, params=None):
        if params is None:
            params = OrderedDict(self.named_parameters())

        in_proj_weight = params.get('in_proj_weight', None)
        in_proj_bias = params.get('in_proj_bias', None)
        bias_k = params.get('bias_k', None)
        bias_v = params.get('bias_v', None)

        if not self._qkv_same_embed_dim:
            return F.multi_head_attention_forward(
                query, key, value, self.embed_dim, self.num_heads,
                in_proj_weight, in_proj_bias,
                bias_k, bias_v, self.add_zero_attn,
                self.dropout, params['out_proj.weight'], params['out_proj.bias'],
                training=self.training,
                key_padding_mask=key_padding_mask, need_weights=need_weights,
                attn_mask=attn_mask, use_separate_proj_weight=True,
                q_proj_weight=params['q_proj_weight'],
                k_proj_weight=params['k_proj_weight'],
                v_proj_weight=params['v_proj_weight'])
        else:
            return F.multi_head_attention_forward(
                query, key, value, self.embed_dim, self.num_heads,
                in_proj_weight, in_proj_bias,
                bias_k, bias_v, self.add_zero_attn,
                self.dropout, params['out_proj.weight'], params['out_proj.bias'],
                training=self.training,
                key_padding_mask=key_padding_mask, need_weights=need_weights,
                attn_mask=attn_mask)



import torch
import torch.nn as nn
import re
import warnings

from collections import OrderedDict


class MetaModule(nn.Module):
    """
    Base class for PyTorch meta-learning modules. These modules accept an
    additional argument `params` in their `forward` method.

    Notes
    -----
    Objects inherited from `MetaModule` are fully compatible with PyTorch
    modules from `torch.nn.Module`. The argument `params` is a dictionary of
    tensors, with full support of the computation graph (for differentiation).
    """
    def __init__(self):
        super(MetaModule, self).__init__()
        self._children_modules_parameters_cache = dict()

    def meta_named_parameters(self, prefix='', recurse=True):
        gen = self._named_members(
            lambda module: module._parameters.items()
            if isinstance(module, MetaModule) else [],
            prefix=prefix, recurse=recurse)
        for elem in gen:
            yield elem

    def meta_parameters(self, recurse=True):
        for name, param in self.meta_named_parameters(recurse=recurse):
            yield param

    def get_subdict(self, params, key=None):
        if params is None:
            return None

        all_names = tuple(params.keys())
        if (key, all_names) not in self._children_modules_parameters_cache:
            if key is None:
                self._children_modules_parameters_cache[(key, all_names)] = all_names

            else:
                key_escape = re.escape(key)
                key_re = re.compile(r'^{0}\.(.+)'.format(key_escape))

                self._children_modules_parameters_cache[(key, all_names)] = [
                    key_re.sub(r'\1', k) for k in all_names if key_re.match(k) is not None]

        names = self._children_modules_parameters_cache[(key, all_names)]
        if not names:
            warnings.warn('Module `{0}` has no parameter corresponding to the '
                          'submodule named `{1}` in the dictionary `params` '
                          'provided as an argument to `forward()`. Using the '
                          'default parameters for this submodule. The list of '
                          'the parameters in `params`: [{2}].'.format(
                          self.__class__.__name__, key, ', '.join(all_names)),
                          stacklevel=2)
            return None

        return OrderedDict([(name, params[f'{key}.{name}']) for name in names])
